from ._discrete_markov_process import (
    banded_square_matrix,
    dmp_marginal_dist,
    dmp_sample_log_likelihood,
    dmp_stationary_dist,
    sample_dmp,
    sample_sparse_ergodic_transition_matrix,
)
from ._hamming_dist import hamming_dist
from ._motif import motif_search
